/*
 * Framebuffer driver for VBE 2.0+ compliant graphic boards.
 * Kernel thread and vm86 routines.
 *
 * (c) 2004-2005 Michal Januszewski <spock@gentoo.org>
 *
 */

#include <linux/config.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/completion.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/signal.h>
#include <linux/suspend.h>
#include <linux/unistd.h>
#include <video/vesa.h>
#include <video/edid.h>
#include <asm/mman.h>
#include <asm/page.h>
#include <asm/vm86.h>
#include <asm/thread_info.h>
#include <asm/uaccess.h>
#include <asm/mmu_context.h>
#include "edid.h"

#ifdef MODULE
int errno;
#endif

static DECLARE_COMPLETION(vesafb_th_completion);
static DECLARE_MUTEX(vesafb_task_list_sem);
static LIST_HEAD(vesafb_task_list);
static DECLARE_WAIT_QUEUE_HEAD(vesafb_wait);

static struct vm86_struct vm86;
static int vesafb_pid = 0;

_syscall3(int,ioperm,unsigned long, a, unsigned long, b, unsigned long, c);
_syscall1(int,vm86old,struct vm86_struct __user*, v86);

#define DEFAULT_VM86_FLAGS (IF_MASK | IOPL_MASK)
#define VM86_PUSHW(x)					\
do { 							\
	vm86.regs.esp -= 2; 				\
	*(u16*)(STACK_ADDR + vm86.regs.esp) = x;	\
} while(0);

/* Stack, the return code and buffers will be put into
 * one contiguous memory chunk:
 *
 * [ STACK | RET_CODE | BUFFER ]
 *
 * We only need a buffer that is ca. 0x2000 bytes in size.
 * Some video BIOSes (sis6326) try to store data somewhere
 * in 0x7000-0x7fff, so we zeromap more memory to be safe.
 */
#define IVTBDA_SIZE 	PAGE_SIZE
#define RET_CODE_SIZE	0x0010
#define STACK_SIZE	0x0500
#define BUFFER_SIZE	0x10000

/* The amount of memory that will be allocated should be a multiple
 * of PAGE_SIZE. */
#define __MEM_SIZE 	(RET_CODE_SIZE + STACK_SIZE + BUFFER_SIZE)
#define REAL_MEM_SIZE	(((__MEM_SIZE / PAGE_SIZE) + 1) * PAGE_SIZE)

#define IVTBDA_ADDR	0x00000
#define STACK_ADDR	(IVTBDA_ADDR + IVTBDA_SIZE)
#define RET_CODE_ADDR	(STACK_ADDR + STACK_SIZE)
#define BUF_ADDR	(RET_CODE_ADDR + RET_CODE_SIZE)

#define FLAG_D 		(1 << 10)

/* Segment prefix opcodes */
enum {
	P_CS = 0x2e,
	P_SS = 0x36,
	P_DS = 0x3e,
	P_ES = 0x26,
	P_FS = 0x64,
	P_GS = 0x65
};

/* Emulated vm86 ins instruction */
static void vm86_ins(int size)
{
	u32 edx, edi;
	edx = vm86.regs.edx & 0xffff;
	edi = (vm86.regs.edi & 0xffff) + (u32)(vm86.regs.es << 4);

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("std\n");
	else
		asm volatile ("cld\n");

	switch (size) {
	case 4:
		asm volatile ("insl\n" : "=D" (edi) : "d" (edx), "0" (edi));
		break;
	case 2:
		asm volatile ("insw\n" : "=D" (edi) : "d" (edx), "0" (edi));
		break;
	case 1:
		asm volatile ("insb\n" : "=D" (edi) : "d" (edx), "0" (edi));
		break;
	}

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("cld\n");

	edi -= (u32)(vm86.regs.es << 4);

	vm86.regs.edi &= 0xffff0000;
	vm86.regs.edi |= edi & 0xffff;
}

static void vm86_rep_ins(int size)
{
	u16 cx = vm86.regs.ecx;
	while (cx--)
		vm86_ins(size);

	vm86.regs.ecx &= 0xffff0000;
}

/* Emulated vm86 outs instruction */
static void vm86_outs(int size, int segment)
{
	u32 edx, esi, base;

	edx = vm86.regs.edx & 0xffff;
	esi = vm86.regs.esi & 0xffff;

	switch (segment) {
	case P_CS: base = vm86.regs.cs; break;
	case P_SS: base = vm86.regs.ss; break;
	case P_ES: base = vm86.regs.es; break;
	case P_FS: base = vm86.regs.fs; break;
	case P_GS: base = vm86.regs.gs; break;
	default:   base = vm86.regs.ds; break;
	}

	esi += base << 4;

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("std\n");
	else
		asm volatile ("cld\n");

	switch (size) {
	case 4:
		asm volatile ("outsl\n" : "=S" (esi) : "d" (edx), "0" (esi));
		break;
	case 2:
		asm volatile ("outsw\n" : "=S" (esi) : "d" (edx), "0" (esi));
		break;
	case 1:
		asm volatile ("outsb\n" : "=S" (esi) : "d" (edx), "0" (esi));
		break;
	}

	if (vm86.regs.eflags & FLAG_D)
		asm volatile ("cld");

	esi -= base << 4;
	vm86.regs.esi &= 0xffff0000;
	vm86.regs.esi |= (esi & 0xffff);
}

static void vm86_rep_outs(int size, int segment)
{
	u16 cx = vm86.regs.ecx;
	while (cx--)
		vm86_outs(size, segment);

	vm86.regs.ecx &= 0xffff0000;
}

static int vm86_do_unknown(void)
{
	u8 data32 = 0, segment = P_DS, rep = 0;
	u8 *instr;
	int ret = 0, i = 0;

	instr = (u8*)((vm86.regs.cs << 4) + vm86.regs.eip);

	while (1) {
		switch(instr[i]) {
		case 0x66:	/* operand size prefix */
			data32 = 1 - data32;
			i++;
			break;
		case 0xf2:	/* repnz */
		case 0xf3:	/* rep */
			rep = 1;
			i++;
			break;
		case P_CS:	/* segment prefix */
		case P_SS:
		case P_DS:
		case P_ES:
		case P_FS:
		case P_GS:
			segment = instr[i];
			i++;
			break;
		case 0xf0:	/* LOCK - ignored */
		case 0x67:	/* address size prefix - ignored */
			i++;
			break;
		case 0x6c:	/* insb */
			if (rep)
				vm86_rep_ins(1);
			else
				vm86_ins(1);
			i++;
			goto out;
		case 0x6d:	/* insw / insd */
			if (rep) {
				if (data32)
					vm86_rep_ins(4);
				else
					vm86_rep_ins(2);
			} else {
				if (data32)
					vm86_ins(4);
				else
					vm86_ins(2);
			}
			i++;
			goto out;
		case 0x6e:	/* outsb */
			if (rep)
				vm86_rep_outs(1, segment);
			else
				vm86_outs(1, segment);
			i++;
			goto out;
		case 0x6f:	/* outsw / outsd */
			if (rep) {
				if (data32)
					vm86_rep_outs(4, segment);
				else
					vm86_rep_outs(2, segment);
			} else {
				if (data32)
					vm86_outs(4, segment);
				else
					vm86_outs(2, segment);
			}
			i++;
			goto out;
		case 0xe4:	/* inb xx */
			asm volatile (
				"inb %w1, %b0"
				: "=a" (vm86.regs.eax)
				: "d" (instr[i+1]), "0" (vm86.regs.eax));
			i += 2;
			goto out;
		case 0xe5:	/* inw xx / ind xx */
			if (data32) {
				asm volatile (
					"inl %w1, %0"
					: "=a" (vm86.regs.eax)
					: "d" (instr[i+1]),
					  "0" (vm86.regs.eax));
			} else {
				asm volatile (
					"inw %w1, %w0"
					: "=a" (vm86.regs.eax)
					: "d" (instr[i+1]),
					  "0" (vm86.regs.eax));
			}
			i += 2;
			goto out;

		case 0xec:	/* inb dx */
			asm volatile (
				"inb %w1, %b0"
	 			: "=a" (vm86.regs.eax)
				: "d" (vm86.regs.edx), "0" (vm86.regs.eax));
			i++;
			goto out;
		case 0xed:	/* inw dx / ind dx */
			if (data32) {
				asm volatile (
					"inl %w1, %0"
					: "=a" (vm86.regs.eax)
					: "d" (vm86.regs.edx));
			} else {
				asm volatile (
					"inw %w1, %w0"
					: "=a" (vm86.regs.eax)
					: "d" (vm86.regs.edx));
			}
			i++;
			goto out;
		case 0xe6:	/* outb xx */
			asm volatile (
				"outb %b0, %w1"
				: /* no return value */
				: "a" (vm86.regs.eax), "d" (instr[i+1]));
			i += 2;
			goto out;
		case 0xe7:	/* outw xx / outd xx */
			if (data32) {
				asm volatile (
					"outl %0, %w1"
					: /* no return value */
					: "a" (vm86.regs.eax),
					  "d" (instr[i+1]));
			} else {
				asm volatile (
					"outw %w0, %w1"
					: /* no return value */
					: "a" (vm86.regs.eax),
					  "d" (instr[i+1]));
			}
			i += 2;
			goto out;
		case 0xee:	/* outb dx */
			asm volatile (
				"outb %b0, %w1"
				: /* no return value */
				: "a" (vm86.regs.eax), "d" (vm86.regs.edx));
			i++;
			goto out;
		case 0xef:	/* outw dx / outd dx */
			if (data32) {
				asm volatile (
					"outl %0, %w1"
					: /* no return value */
					: "a" (vm86.regs.eax),
					  "d" (vm86.regs.edx));
			} else {
				asm volatile (
					"outw %w0, %w1"
					: /* no return value */
					: "a" (vm86.regs.eax),
					  "d" (vm86.regs.edx));
			}
			i++;
			goto out;
		default:
			printk(KERN_ERR "vesafb: BUG, opcode 0x%x emulation "
					"not supported (EIP: 0x%lx)\n",
					instr[i], (u32)(vm86.regs.cs << 4) +
					vm86.regs.eip);
			ret = 1;
			goto out;
		}
	}
out: 	vm86.regs.eip += i;
	return ret;
}

void vesafb_do_vm86(struct vm86_regs *regs)
{
	unsigned int ret;
	u8 *retcode = (void*)RET_CODE_ADDR;

	memset(&vm86,0,sizeof(vm86));
	memcpy(&vm86.regs, regs, sizeof(struct vm86_regs));

	/* The return code */
	retcode[0] = 0xcd;  		/* int opcode */
	retcode[1] = 0xff;		/* int number (255) */

        /* We use int 0xff to get back to protected mode */
	memset(&vm86.int_revectored, 0, sizeof(vm86.int_revectored));
        ((unsigned char *)&vm86.int_revectored)[0xff / 8] |= (1 << (0xff % 8));

	/*
	 * We want to call int 0x10, so we set:
	 *   CS = 0x42 = 0x10 * 4 + 2
	 *   IP = 0x40 = 0x10 * 4
	 * and SS:ESP. It's up to the caller to set the rest of the registers.
	 */
	vm86.regs.eflags = DEFAULT_VM86_FLAGS;
	vm86.regs.cs = *(unsigned short *)0x42;
	vm86.regs.eip = *(unsigned short *)0x40;
	vm86.regs.ss = (STACK_ADDR >> 4);
	vm86.regs.esp = ((STACK_ADDR & 0x0000f) + STACK_SIZE);

	/* These will be fetched off the stack when we come to an iret in the
	 * int's 0x10 code. */
	VM86_PUSHW(DEFAULT_VM86_FLAGS);
	VM86_PUSHW((RET_CODE_ADDR >> 4));	/* return code segment */
	VM86_PUSHW((RET_CODE_ADDR & 0x0000f));	/* return code offset */

	while(1) {
		ret = vm86old(&vm86);

		if (VM86_TYPE(ret) == VM86_INTx) {
			int vint = VM86_ARG(ret);

			/* If exit from vm86 was caused by int 0xff, then
			 * we're done.. */
			if (vint == 0xff)
				goto out;

			/* .. otherwise, we have to call the int handler
			 * manually */
			VM86_PUSHW(vm86.regs.eflags);
			VM86_PUSHW(vm86.regs.cs);
			VM86_PUSHW(vm86.regs.eip);

			vm86.regs.cs = *(u16 *)((vint << 2) + 2);
			vm86.regs.eip = *(u16 *)(vint << 2);
			vm86.regs.eflags &= ~(VIF_MASK | TF_MASK);
		} else if (VM86_TYPE(ret) == VM86_UNKNOWN) {
			if (vm86_do_unknown())
				goto out;
		} else {
			printk(KERN_ERR "vesafb: BUG, returned from "
					"vm86 with %x (EIP: 0x%lx)\n",
					ret, (u32)(vm86.regs.cs << 4) +
					vm86.regs.eip);
			goto out;
		}
	}

out:	/* copy the registers' state back to the caller's struct */
	memcpy(regs, &vm86.regs, sizeof(struct vm86_regs));
}

static int vesafb_remap_pfn_range(unsigned long start, unsigned long end,
				  unsigned long pgoff, unsigned long prot,
				  int type)
{
	struct vm_area_struct *vma;
	struct mm_struct *mm = current->mm;
	int ret = 0;

	vma = kmem_cache_alloc(vm_area_cachep, SLAB_KERNEL);
	if (!vma)
		return -ENOMEM;
	memset(vma, 0, sizeof(*vma));
	down_write(&mm->mmap_sem);
	vma->vm_mm = mm;
	vma->vm_start = start;
	vma->vm_end = end;
	vma->vm_flags = VM_READ | VM_WRITE | VM_EXEC;
	vma->vm_flags |= mm->def_flags;
	vma->vm_page_prot.pgprot = prot;
	vma->vm_pgoff = pgoff;

	if ((ret = insert_vm_struct(mm, vma))) {
		up_write(&mm->mmap_sem);
		kmem_cache_free(vm_area_cachep, vma);
		return ret;
	}

	if (type) {
		vma->vm_flags |= VM_SHARED;
		ret = zeromap_page_range(vma,
					 vma->vm_start,
					 vma->vm_end - vma->vm_start,
					 vma->vm_page_prot);
	} else {
		ret = remap_pfn_range(vma,
				      vma->vm_start,
				      vma->vm_pgoff,
				      vma->vm_end - vma->vm_start,
				      vma->vm_page_prot);
	}
	up_write(&mm->mmap_sem);
	return ret;
}

static inline int vesafb_init_mem(void)
{
	int ret = 0;

	/* The memory chunks we're remapping here should be multiples
	 * of PAGE_SIZE. */
	ret += vesafb_remap_pfn_range(0x00000, IVTBDA_SIZE, 0,
				      PROT_READ | PROT_EXEC | PROT_WRITE, 0);
	ret += vesafb_remap_pfn_range(IVTBDA_SIZE, REAL_MEM_SIZE, 0,
				      PROT_READ | PROT_EXEC | PROT_WRITE, 1);
	ret += vesafb_remap_pfn_range(0x9f000, 0x100000,
				      0x9f000 >> PAGE_SHIFT,
				      PROT_READ | PROT_EXEC | PROT_WRITE, 0);
	if (ret)
		printk(KERN_ERR "vesafb thread: memory remapping failed\n");

	return ret;
}

#define vesafb_get_string(str) \
{ 									\
	/* The address is in the form ssssoooo, where oooo = offset,	\
	 * ssss = segment */						\
	addr = ((p_vbe(tsk->buf)->str & 0xffff0000) >> 12) +		\
		(p_vbe(tsk->buf)->str & 0x0000ffff);			\
									\
	/* The data is in ROM which is shared between processes, so we 	\
	 * just translate the real mode address into one visible from 	\
	 * kernel space */						\
	if (addr >= 0xa0000) {						\
		p_vbe(tsk->buf)->str = (u32) __va(addr);		\
									\
	/* The data is in the buffer, we just have to convert the	\
	 * address so that it points into the buffer user provided. */	\
	} else if (addr > BUF_ADDR && addr < BUF_ADDR +			\
		   sizeof(struct vesafb_vbe_ib)) {			\
		addr -= BUF_ADDR;					\
		p_vbe(tsk->buf)->str = (u32) (tsk->buf + addr);		\
									\
	/* This should never happen: someone was insane enough to put	\
	 * the data somewhere in RAM.. */				\
	} else {							\
		p_vbe(tsk->buf)->str = (u32) "";			\
	}								\
}

void vesafb_handle_getvbeib(struct vesafb_task *tsk)
{
	int addr, res;

	tsk->regs.es  = (BUF_ADDR >> 4);
	tsk->regs.edi = (BUF_ADDR & 0x000f);
	strncpy(p_vbe(BUF_ADDR)->vbe_signature, "VBE2", 4);

	vesafb_do_vm86(&tsk->regs);
	memcpy(tsk->buf, (void*)(BUF_ADDR), sizeof(struct vesafb_vbe_ib));

	/* The OEM fields were not defined prior to VBE 2.0 */
	if (p_vbe(tsk->buf)->vbe_version >= 0x200) {
		vesafb_get_string(oem_string_ptr);
		vesafb_get_string(oem_vendor_name_ptr);
		vesafb_get_string(oem_product_name_ptr);
		vesafb_get_string(oem_product_rev_ptr);
	}

	/* This is basically the same as vesafb_get_string() */
	addr = ((p_vbe(tsk->buf)->mode_list_ptr & 0xffff0000) >> 12) +
		(p_vbe(tsk->buf)->mode_list_ptr & 0x0000ffff);

	if (addr >= 0xa0000) {
		p_vbe(tsk->buf)->mode_list_ptr = (u32) __va(addr);
	} else if (addr > BUF_ADDR && addr < BUF_ADDR +
		   sizeof(struct vesafb_vbe_ib)) {
		addr -= BUF_ADDR;
		p_vbe(tsk->buf)->mode_list_ptr = (u32) (tsk->buf + addr);
	} else {
		res = 0;
		printk(KERN_WARNING "vesafb: warning, copying modelist "
				    "from somewhere in RAM!\n");
		while (*(u16*)(addr+res) != 0xffff &&
		       res < (sizeof(p_vbe(tsk->buf)->reserved) - 2)) {
			*(u16*) ((u32)&(p_vbe(tsk->buf)->reserved) + res) =
				*(u16*)(addr+res);
			res += 2;
		}
		*(u16*) ((u32)&(p_vbe(tsk->buf)->reserved) + res) = 0xffff;
	}
}

int vesafb_handle_tasks(void)
{
	struct vesafb_task *tsk;
	struct list_head *curr, *next;
	int ret = 0;

	down(&vesafb_task_list_sem);
	list_for_each_safe(curr, next, &vesafb_task_list) {
		tsk = list_entry(curr, struct vesafb_task, node);

		if (tsk->flags & TF_EXIT) {
			ret = 1;
			goto task_done;
		}
		if (tsk->flags & TF_GETVBEIB) {
			vesafb_handle_getvbeib(tsk);
			goto task_done;
		}
		if (tsk->flags & TF_MASK_BUF) {
			tsk->regs.es  = (BUF_ADDR >> 4);
			tsk->regs.edi = (BUF_ADDR & 0x000f);
		}
		if (tsk->flags & TF_USE_BUF)
			memcpy((void*)BUF_ADDR, tsk->buf, tsk->buf_len);

		vesafb_do_vm86(&tsk->regs);

		if (tsk->flags & TF_RETURN_BUF)
			memcpy(tsk->buf, (void*)BUF_ADDR, tsk->buf_len);

task_done:	list_del(curr);
		complete(&tsk->done);
	}

	/* If we're going to kill this thread, don't allow any elements
	 * to be added to the task list. */
	if (!ret)
		up(&vesafb_task_list_sem);

	return ret;
}

/*
 * This 'hybrid' thread serves as a backend for vesafb-tng, handling all vm86
 * calls. It is started as a kernel thread. It then creates its own mm struct,
 * thus separating itself from any userspace processes. At this moment, it
 * stops being a kernel thread (kernel threads have mm = NULL) and becomes
 * a 'hybrid' thread -- one that has full access to kernel space, yet runs
 * with its own address space.
 *
 * This is necessary because in order to make vm86 calls some parts of the
 * first 1MB of RAM have to be setup to mimic the real mode. These are:
 *  - interrupt vector table	[0x00000-0x003ff]
 *  - BIOS data area		[0x00400-0x004ff]
 *  - Extended BIOS data area	[0x9fc00-0x9ffff]
 *  - the video RAM		[0xa0000-0xbffff]
 *  - video BIOS		[0xc0000-0xcffff]
 *  - motherboard BIOS		[0xf0000-0xfffff]
 */
int vesafb_thread(void *unused)
{
	int err = 0;

	set_fs(KERNEL_DS);
	daemonize("vesafb");

	if (set_new_mm()) {
		err = -ENOMEM;
		goto thr_end;
	}
	if (vesafb_init_mem()) {
		err = -ENOMEM;
		goto thr_end;
	}

	DPRINTK("started vesafb thread\n");

	/* Having an IO bitmap makes things faster as we avoid GPFs
	 * when running vm86 code. We can live if it fails, though,
	 * so don't bother checking for errors. */
	ioperm(0,1024,1);
	set_user_nice(current, -10);

	complete(&vesafb_th_completion);

	while (1) {
		if (vesafb_handle_tasks())
			break;
		wait_event_interruptible(vesafb_wait,
					 !list_empty(&vesafb_task_list));
		try_to_freeze();
	}

out:	DPRINTK("exiting the vesafb thread\n");
	vesafb_pid = -1;

	/* Now that all callers know this thread is no longer running
	 * (pid < 0), allow them to continue. */
	up(&vesafb_task_list_sem);
	return err;
thr_end:
	down(&vesafb_task_list_sem);
	complete(&vesafb_th_completion);
	goto out;
}

int vesafb_queue_task(struct vesafb_task *tsk)
{
	down(&vesafb_task_list_sem);
	if (vesafb_pid < 0)
		return -1;
	list_add_tail(&tsk->node, &vesafb_task_list);
	up(&vesafb_task_list_sem);
	wake_up(&vesafb_wait);
	return 0;
}

int vesafb_wait_for_thread(void)
{
	/* PID 0 means that the thread is still initializing. */
	if (vesafb_pid < 0)
		return -1;
	wait_for_completion(&vesafb_th_completion);
	return 0;
}

int __init vesafb_init_thread(void)
{
	vesafb_pid = kernel_thread(vesafb_thread,NULL,0);
	return 0;
}

#ifdef MODULE
void __exit vesafb_kill_thread(void)
{
	struct vesafb_task *tsk;
	if (vesafb_pid <= 0)
		return;

	vesafb_create_task(tsk);
	if (!tsk)
		return;
	tsk->flags |= TF_EXIT;
	vesafb_queue_task(tsk);
	vesafb_wait_for_task(tsk);
	kfree(tsk);
	return;
}
module_exit(vesafb_kill_thread);
#endif
module_init(vesafb_init_thread);

EXPORT_SYMBOL_GPL(vesafb_queue_task);
EXPORT_SYMBOL_GPL(vesafb_wait_for_thread);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Michal Januszewski");

